﻿// RepRap NoLash Software BackLash Compensation Tool
// Written by Richard White on Jan. 22, 2013
// email georgeflug@gmail.com

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;

namespace WindowsFormsApplication1
{
    public partial class lash : Form
    {
        public lash()
        {
            InitializeComponent();
        }

        private void button1_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == System.Windows.Forms.DialogResult.OK)
            {
                txtFile.Text = ofd.FileName;
            }
        }

        private void button2_Click(object sender, EventArgs e)
        {

            // verify file exists
            if (!File.Exists(txtFile.Text))
            {
                MessageBox.Show("File does not exist.");
                return;
            }

            // declare state variables
            double xPos = 0;
            double yPos = 0;
            int xDir = 1;
            int yDir = 1;

            // save the amount of lash compensation
            double xLash = double.Parse(txtX.Text) / 2; // divide by 2 to compensate for 2 directions of movement
            double yLash = double.Parse(txtY.Text) / 2; // (forward and backward)

            // open source file and open temporary file
            StreamReader fileLash = new StreamReader(txtFile.Text);
            StreamWriter fileNew;
            if (txtFile.Text.IndexOf("_export") > -1)
            {
                fileNew = new StreamWriter(txtFile.Text.Replace("_export", "_nolash"));
            }
            else
            {
                string strNew = Path.GetDirectoryName(txtFile.Text) +
                    Path.DirectorySeparatorChar + 
                    Path.GetFileNameWithoutExtension(txtFile.Text) + "_nolash" +
                    Path.GetExtension(txtFile.Text);
                fileNew = new StreamWriter(strNew);
            }

            // loop through every line
            while (!fileLash.EndOfStream)
            {

                // get the line of text
                string strLine = fileLash.ReadLine();
                
                // get the gcode command
                int intIndex = strLine.IndexOf(" ");
                string strCommand;
                if (intIndex == -1)
                {
                    strCommand = strLine;
                }
                else
                {
                    strCommand = strLine.Remove(strLine.IndexOf(" "));
                }

                // compare commands
                switch (strCommand.ToUpper())
                {
                    case "G0":
                    case "G1":
                       
                        // recalculate x position
                        string strCountermeasure = string.Empty;
                        xPos = repairCommand(xPos, ref xDir, "X", xLash, ref strLine, ref strCountermeasure);

                        // recalculate y position
                        yPos = repairCommand(yPos, ref yDir, "Y", yLash, ref strLine, ref strCountermeasure);

                        // print the countermeasure string to the new file
                        if (strCountermeasure.Length > 0)
                        {
                            fileNew.WriteLine(strCountermeasure);
                        }

                        break;

                    case "G28":

                        // save new positions for x and y
                        //int intCmd = strCommand.IndexOf("X");
                        //int intCmdEnd = strCommand.IndexOf(" ", intCmd);
                        //xPos = double.Parse(strCommand.Substring(intCmd + 1, intCmdEnd - intCmd - 1));
                        //intCmd = strCommand.IndexOf("Y");
                        //intCmdEnd = strCommand.IndexOf(" ", intCmd);
                        //xPos = double.Parse(strCommand.Substring(intCmd + 1, intCmdEnd - intCmd - 1));
                        xPos = 0;
                        yPos = 0;
                        xDir = 1;
                        yDir = 1;

                        break;
                    default:
                        break;
                }

                // save the edited gcode
                fileNew.WriteLine(strLine);

            }

            // close both files
            fileLash.Close();
            fileNew.Close();

            // replace the original file
            //File.Delete(txtFile.Text);
            //File.Move(txtFile.Text + "_new", txtFile.Text);

            // indicate completion
            MessageBox.Show("Done");

        }

        private double repairCommand(double currentPos, ref int currentDir, string axis, double lash, ref string command, ref string countermeasure)
        {

            // this routine returns the new position in the X or Y direction (depending of the axis parameter)
            // axis should be "X" or "Y"

            // get the location of the position command
            int intCmd = command.IndexOf(axis);

            // if the axis does not move during this command then do nothing
            if (intCmd == -1)
            {
                return currentPos;
            }

            // get the ending location of the position command (denoted by a space or a semicolon)
            int intCmdEnd = command.IndexOf(" ", intCmd);
            if (intCmdEnd == -1)
            {
                intCmdEnd = command.IndexOf(";", intCmd);
            }
            else
            {
                int intComment = command.IndexOf(";", intCmd);
                if (intComment > -1) intCmdEnd = Math.Min(intCmdEnd, intComment);
            }
            if (intCmdEnd == -1) intCmdEnd = command.Length;
            
            // parse the command
            string strPos = command.Substring(intCmd + 1, intCmdEnd - intCmd - 1);
            double newPos = double.Parse(strPos);

            // recalculated position
            double recalc;

            // compare to currect position, implement lash countermeasure if necessary
            if (comparePositions(currentPos, newPos, currentDir) == false)
            {
                currentDir = -currentDir; // reverse the direction of movement
                double calc = currentPos + lash * currentDir; // calculate countermeasure position
                // add to 'countermeasure' string
                if (countermeasure.Length == 0)
                {
                    countermeasure = "G1";
                }
                countermeasure += " " + axis + calc.ToString();

            }
            recalc = newPos + lash * currentDir; // save recalculated position

            // update command
            command = command.Remove(intCmd) + axis + recalc.ToString() + command.Substring(intCmdEnd);
        
            // update position
            return newPos;

        }

        private bool comparePositions(double currentPos, double newPos, int currentDir)
        {

            // this routine compares the two positions and determines the direction of movement.
            // then it compares the direction of movement to the previous direction of movement.
            // it returns true if the direction is the same, it returns false otherwise

            int newDir = 0;

            if (currentPos < newPos)
            {
                newDir = 1;
            }
            else if (currentPos > newPos)
            {
                newDir = -1;
            }
            else
            {
                return true;
            }

            return (newDir == currentDir);

        }
    }
}
